@@ -1,51 +1,128 @@
 from abc import ABC as _ABC, abstractmethod as _abstractmethod
 from functools import reduce as _reduce
-
-from .utils import *
-from .utils import (
-    _crypto,
-    _NoEncryption,
-    _TextEncryption,
-    _Encoding,
-    _PrivateFormat,
-    _pkcs12,
-    _X509Creds,
-    _X509PubCreds,
+from typing import (
+    TYPE_CHECKING,
+    overload,
+    Literal as _Literal,
+    Iterable as _Iter,
+    Mapping as _Mapping,
 )
+
 import tempfile as _tmp
 import os as _os
 
+from ._vendor.pyopenssl import get_pyopenssl_ctx, _crypto
+from ._vendor.crypto import *
+from .encoding import (
+    Encoding,
+    X509EncodedStore,
+    pem as _pem,
+    der as _der,
+    pkcs12 as _pkcs12,
+    _Encoding,
+    Encoded,
+    PasswordLike
+)
+from .encoding._decoders import ValidPath
+from ._models import _X509Credentials, _X509Identity
+from .utils import (
+    CertPurpose,
+    ExtensionLike,
+    KeyUsage,
+    cert_builder,
+    generate_certificate,
+    is_public_key,
+    DatetimeRef,
+    _IPAddress as IPAddress
+)
+
+if TYPE_CHECKING:
+    from sslcontext import SSLContext
+
+
+class X509Issuer(_ABC):
+    @_abstractmethod
+    def sign(
+        self, builder: CertificateBuilder, hash_alg: HashAlgorithm
+    ) -> "X509Identity":
+        ...
 
-class X509PublicCredentials(_X509PubCreds):
     @overload
-    def dump(self, encoding: Literal[Encoding.PEM, Encoding.PKCS12]) -> bytes:
+    def generate(
+        self,
+        subject: "x509.Name|str",
+        key: "PrivateKey|int|None" = None,
+        purpose: CertPurpose = None,
+        not_before: DatetimeRef = None,
+        not_after: DatetimeRef = None,
+        extensions: _Iter[ExtensionLike] = None,
+        key_usage: "_Mapping[KeyUsage,bool]" = None,
+        ext_key_usage: "list" = None,
+        hash_alg: HashAlgorithm = None,
+    ) -> "X509Credentials":
+        ...
+
+    @overload
+    def generate(
+        self,
+        subject: "x509.Name|str",
+        key: "PublicKey" = None,
+        purpose: CertPurpose = None,
+        not_before: DatetimeRef = None,
+        not_after: DatetimeRef = None,
+        extensions: "_Iter[ExtensionLike]" = None,
+        key_usage: "_Mapping[KeyUsage,bool]" = None,
+        ext_key_usage: "list" = None,
+        hash_alg: HashAlgorithm = None,
+    ) -> "X509Identity":
+        ...
+
+    def generate(
+        self,
+        subject: "x509.Name|str",
+        key: "PrivateKey|PublicKey|int|None" = None,
+        purpose: CertPurpose = None,
+        not_before: DatetimeRef = None,
+        not_after: DatetimeRef = None,
+        extensions: _Iter[ExtensionLike] = None,
+        key_usage: "_Mapping[KeyUsage,bool]" = None,
+        ext_key_usage: "list" = None,
+        hash_alg: HashAlgorithm = None,
+    ):
+        return create_creds(
+            subject,
+            key,
+            self,
+            purpose,
+            not_before,
+            not_after,
+            extensions,
+            key_usage,
+            ext_key_usage,
+            hash_alg,
+        )
+
+
+class X509Identity(_X509Identity):
+    @overload
+    def dump(self, encoding: _Literal[Encoding.PEM, Encoding.PKCS12]) -> bytes:
         pass
 
     @overload
-    def dump(self, encoding: Literal[Encoding.DER]) -> "tuple[bytes, list[bytes]]":
+    def dump(self, encoding: _Literal[Encoding.DER]) -> "tuple[bytes, list[bytes]]":
         pass
 
     def dump(self, encoding: Encoding):
-
+        encoding
         if encoding is Encoding.DER:
             return (
                 self.cert.public_bytes(_Encoding.DER),
                 [ca.public_bytes(_Encoding.DER) for ca in self.chain],
             )
         elif encoding is Encoding.PEM:
-            return _reduce(
-                lambda acc, cert: acc + cert.public_bytes(_Encoding.PEM),
-                self.chain,
-                self.cert.public_bytes(_Encoding.PEM),
-            )
+            return _pem.dump(None, *self)
         elif encoding is Encoding.PKCS12:
-            return _pkcs12.serialize_key_and_certificates(
-                self.cert.subject.rfc4514_string(),
-                None,
-                self.cert,
-                self.chain,
-                _NoEncryption(),
-            )
+            return _pkcs12.dump(None, *self)
         else:
             raise ValueError(f"Invalid encoding {encoding}")
 
@@ -63,10 +140,10 @@ class X509PublicCredentials(_X509PubCreds):
         key: "PublicKey|PrivateKey" = None,
         issuer: "X509Issuer|None" = None,
         purpose: CertPurpose = None,
-        not_before: "datetime|int|timedelta" = None,
-        not_after: "datetime|int|timedelta" = None,
-        extensions: Iterable[ExtensionLike] = None,
-        key_usage: "dict[KeyUsage,bool]" = None,
+        not_before: DatetimeRef = None,
+        not_after: DatetimeRef = None,
+        extensions: _Iter[ExtensionLike] = None,
+        key_usage: "_Mapping[KeyUsage,bool]" = None,
         ext_key_usage: "list" = None,
         hash_alg: HashAlgorithm = None,
     ):
@@ -88,11 +165,13 @@ class X509PublicCredentials(_X509PubCreds):
             return creds
 
     def apply_to_sslcontext(self, sslcontext: "SSLContext"):
-
-        if is_pyopenssl(sslcontext):
+        pyopenssl = get_pyopenssl_ctx(sslcontext)
+        if pyopenssl:
             crt, chain = self.to_pyopenssl()
-            for ca in [crt, *chain]:
-                sslcontext._ctx.add_extra_chain_cert(ca)
+            store = pyopenssl.get_cert_store()
+            store.add_cert(crt)
+            for ca in chain:
+                store.add_cert(ca)
             return
 
         pem = self.dump(Encoding.PEM)
@@ -115,113 +194,32 @@ class X509PublicCredentials(_X509PubCreds):
         @staticmethod
         def from_pyopenssl(
             cert: _crypto.X509,
-            chain: Iterable[_crypto.X509] = [],
+            chain: _Iter[_crypto.X509] = [],
         ):
             return from_pyopenssl(cert, chain)
 
 
-class X509Issuer(_ABC):
-    @_abstractmethod
-    def sign(
-        self, builder: CertificateBuilder, hash_alg: HashAlgorithm
-    ) -> X509PublicCredentials:
-        ...
-
-    @overload
-    def generate(
-        self,
-        subject: "x509.Name|str",
-        key: "PrivateKey|int|None" = None,
-        purpose: CertPurpose = None,
-        not_before: "datetime|int|timedelta" = None,
-        not_after: "datetime|int|timedelta" = None,
-        extensions: Iterable[ExtensionLike] = None,
-        key_usage: "dict[KeyUsage,bool]" = None,
-        ext_key_usage: "list" = None,
-        hash_alg: HashAlgorithm = None,
-    ) -> "X509Credentials":
-        ...
-
-    @overload
-    def generate(
-        self,
-        subject: "x509.Name|str",
-        key: "PublicKey" = None,
-        purpose: CertPurpose = None,
-        not_before: "datetime|int|timedelta" = None,
-        not_after: "datetime|int|timedelta" = None,
-        extensions: "Iterable[ExtensionLike]" = None,
-        key_usage: "dict[KeyUsage,bool]" = None,
-        ext_key_usage: "list" = None,
-        hash_alg: HashAlgorithm = None,
-    ) -> "X509PublicCredentials":
-        ...
-
-    def generate(
-        self,
-        subject: "x509.Name|str",
-        key: "PrivateKey|PublicKey|int|None" = None,
-        purpose: CertPurpose = None,
-        not_before: "datetime|int|timedelta" = None,
-        not_after: "datetime|int|timedelta" = None,
-        extensions: Iterable[ExtensionLike] = None,
-        key_usage: "dict[KeyUsage,bool]" = None,
-        ext_key_usage: "list" = None,
-        hash_alg: HashAlgorithm = None,
-    ):
-        return create_creds(
-            subject,
-            key,
-            self,
-            purpose,
-            not_before,
-            not_after,
-            extensions,
-            key_usage,
-            ext_key_usage,
-            hash_alg,
-        )
-
-
-class X509Credentials(_X509Creds, X509Issuer):
+class X509Credentials(_X509Credentials, X509Issuer):
     @overload
     def dump(
-        self, encoding: Literal[Encoding.PEM, Encoding.PKCS12], password: str = None
+        self, encoding: _Literal[Encoding.PEM, Encoding.PKCS12], password: str = None
     ) -> bytes:
         pass
 
     @overload
     def dump(
-        self, encoding: Literal[Encoding.DER], password: str = None
+        self, encoding: _Literal[Encoding.DER], password: str = None
     ) -> "tuple[bytes, bytes, list[bytes]]":
         pass
 
     def dump(self, encoding: Encoding, password: str = None):
-        encryption = (
-            _NoEncryption() if password is None else _TextEncryption(password.encode())
-        )
+
         if encoding is Encoding.DER:
-            return (
-                self.key.private_bytes(_Encoding.DER, _PrivateFormat.PKCS8, encryption),
-                self.cert.public_bytes(_Encoding.DER),
-                [ca.public_bytes(_Encoding.DER) for ca in self.chain],
-            )
+            return _der.dump(*self, password)
         elif encoding is Encoding.PEM:
-            return self.key.private_bytes(
-                _Encoding.PEM, _PrivateFormat.PKCS8, encryption
-            ) + _reduce(
-                lambda acc, cert: acc + cert.public_bytes(_Encoding.PEM),
-                self.chain,
-                self.cert.public_bytes(_Encoding.PEM),
-            )
+            return _pem.dump(*self, password)
         elif encoding is Encoding.PKCS12:
-            return _pkcs12.serialize_key_and_certificates(
-                self.cert.subject.rfc4514_string(),
-                self.key,
-                self.cert,
-                self.chain,
-                encryption,
-            )
+            return _pkcs12.dump(*self, password)
         else:
             raise ValueError(f"Invalid encoding {encoding}")
 
@@ -235,15 +233,15 @@ class X509Credentials(_X509Creds, X509Issuer):
 
     def sign(
         self, builder: CertificateBuilder, hash_alg: HashAlgorithm = None
-    ) -> X509PublicCredentials:
-        return X509PublicCredentials(
+    ) -> X509Identity:
+        return X509Identity(
             generate_certificate(builder, self, hash_alg),
             [self.cert, *self.chain],
         )
 
     @property
     def public(self):
-        return X509PublicCredentials(self.cert, self.chain)
+        return X509Identity(self.cert, self.chain)
 
     @staticmethod
     def create(
@@ -251,10 +249,10 @@ class X509Credentials(_X509Creds, X509Issuer):
         key: "PrivateKey|int|None" = None,
         issuer: "X509Issuer|None" = None,
         purpose: CertPurpose = None,
-        not_before: "datetime|int|timedelta" = None,
-        not_after: "datetime|int|timedelta" = None,
-        extensions: Iterable[ExtensionLike] = None,
-        key_usage: "dict[KeyUsage,bool]" = None,
+        not_before: DatetimeRef = None,
+        not_after: DatetimeRef = None,
+        extensions: _Iter[ExtensionLike] = None,
+        key_usage: "_Mapping[KeyUsage,bool]" = None,
         ext_key_usage: "list" = None,
         hash_alg: HashAlgorithm = None,
     ):
@@ -274,12 +272,15 @@ class X509Credentials(_X509Creds, X509Issuer):
         )
 
     def apply_to_sslcontext(self, sslcontext: "SSLContext"):
-        if is_pyopenssl(sslcontext):
+        pyopenssl = get_pyopenssl_ctx(sslcontext)
+        if cert:
             key, cert, chain = self.to_pyopenssl()
+            store = pyopenssl.get_cert_store()
+            store.add_cert(cert)
             for ca in chain:
-                sslcontext._ctx.add_extra_chain_cert(ca)
-            sslcontext._ctx.use_certificate(cert)
-            sslcontext._ctx.use_privatekey(key)
+                store.add_cert(ca)
+            pyopenssl.use_certificate(cert)
+            pyopenssl.use_privatekey(key)
             return
 
         pem = self.dump(Encoding.PEM)
@@ -294,9 +295,8 @@ class X509Credentials(_X509Creds, X509Issuer):
     if _crypto:
 
         def to_pyopenssl(self):
-            key = dump_key(self.key, Encoding.PEM)
             return (
-                _crypto.load_privatekey(_crypto.FILETYPE_PEM, key),
+                _crypto.load_privatekey(_crypto.FILETYPE_PEM, _pem.dump_key(self.key)),
                 _crypto.X509.from_cryptography(self.cert),
                 [_crypto.X509.from_cryptography(ca) for ca in self.chain],
             )
@@ -305,7 +305,7 @@ class X509Credentials(_X509Creds, X509Issuer):
         def from_pyopenssl(
             key: _crypto.PKey,
             cert: _crypto.X509,
-            chain: Iterable[_crypto.X509] = [],
+            chain: _Iter[_crypto.X509] = [],
         ):
             return from_pyopenssl(key, cert, chain)
 
@@ -314,26 +314,26 @@ if _crypto:
 
     @overload
     def from_pyopenssl(
-        cert: _crypto.X509, chain: Iterable[_crypto.X509] = [], /
-    ) -> X509PublicCredentials:
+        cert: _crypto.X509, chain: _Iter[_crypto.X509] = [], /
+    ) -> X509Identity:
         ...
 
     @overload
     def from_pyopenssl(
-        key: None, cert: _crypto.X509, chain: Iterable[_crypto.X509] = [], /
-    ) -> X509PublicCredentials:
+        key: None, cert: _crypto.X509, chain: _Iter[_crypto.X509] = [], /
+    ) -> X509Identity:
         ...
 
     @overload
     def from_pyopenssl(
-        key: _crypto.PKey, cert: _crypto.X509, chain: Iterable[_crypto.X509] = [], /
+        key: _crypto.PKey, cert: _crypto.X509, chain: _Iter[_crypto.X509] = [], /
     ) -> X509Credentials:
         ...
 
     def from_pyopenssl(*args):
         cert: _crypto.X509
         key: "None|_crypto.PKey"
-        chain: Iterable[_crypto.X509]
+        chain: _Iter[_crypto.X509]
         if isinstance(args[0], _crypto.X509):
             cert, chain = args
             key = None
@@ -346,7 +346,7 @@ if _crypto:
                 [ca.to_cryptography() for ca in (chain or [])],
             )
             if key
-            else X509PublicCredentials(
+            else X509Identity(
                 cert.to_cryptography(), [ca.to_cryptography() for ca in (chain or [])]
             )
         )
@@ -357,21 +357,19 @@ def load_creds(*stores: Encoded):
     cert = None
     chain = []
     for store in stores:
-        for decoded in X509EncodedStore(store).decode():
+        for decoded in X509EncodedStore(store):
             if isinstance(decoded, Certificate):
                 if cert is None:
                     cert = decoded
                 else:
-                    chain.append(cert)
+                    chain.append(decoded)
             else:
                 if key is None:
                     key = decoded
 
     if cert is None:
         raise ValueError("No certificate was found")
-    return (
-        X509Credentials(key, cert, chain) if key else X509PublicCredentials(cert, chain)
-    )
+    return X509Credentials(key, cert, chain) if key else X509Identity(cert, chain)
 
 
 @overload
@@ -380,10 +378,10 @@ def create_creds(
     key: "PrivateKey|int|None" = None,
     issuer: "X509Issuer|None" = None,
     purpose: CertPurpose = None,
-    not_before: "datetime|int|timedelta" = None,
-    not_after: "datetime|int|timedelta" = None,
-    extensions: Iterable[ExtensionLike] = None,
-    key_usage: "dict[KeyUsage,bool]" = None,
+    not_before: DatetimeRef = None,
+    not_after: DatetimeRef = None,
+    extensions: _Iter[ExtensionLike] = None,
+    key_usage: "_Mapping[KeyUsage,bool]" = None,
     ext_key_usage: "list" = None,
     hash_alg: HashAlgorithm = None,
 ) -> X509Credentials:
@@ -396,13 +394,13 @@ def create_creds(
     key: "PublicKey" = None,
     issuer: "X509Issuer|None" = None,
     purpose: CertPurpose = None,
-    not_before: "datetime|int|timedelta" = None,
-    not_after: "datetime|int|timedelta" = None,
-    extensions: Iterable[ExtensionLike] = None,
-    key_usage: "dict[KeyUsage,bool]" = None,
+    not_before: DatetimeRef = None,
+    not_after: DatetimeRef = None,
+    extensions: _Iter[ExtensionLike] = None,
+    key_usage: "_Mapping[KeyUsage,bool]" = None,
     ext_key_usage: "list" = None,
     hash_alg: HashAlgorithm = None,
-) -> X509PublicCredentials:
+) -> X509Identity:
     ...
 
 
@@ -411,10 +409,10 @@ def create_creds(
     key: "PrivateKey|PublicKey|int|None" = None,
     issuer: "X509Issuer|None" = None,
     purpose: CertPurpose = None,
-    not_before: "datetime|int|timedelta" = None,
-    not_after: "datetime|int|timedelta" = None,
-    extensions: Iterable[ExtensionLike] = None,
-    key_usage: "dict[KeyUsage,bool]" = None,
+    not_before: DatetimeRef = None,
+    not_after: DatetimeRef = None,
+    extensions: _Iter[ExtensionLike] = None,
+    key_usage: "_Mapping[KeyUsage,bool]" = None,
     ext_key_usage: "list" = None,
     hash_alg: HashAlgorithm = None,
 ):
@@ -437,6 +435,6 @@ def create_creds(
         cert, chain = issuer.sign(builder, hash_alg)
 
     if is_public_key(key):
-        return X509PublicCredentials(cert, chain)
+        return X509Identity(cert, chain)
     else:
         return X509Credentials(key, cert, chain)
